# -*- coding: utf-8 -*-
from .meta import *


__all__ = ["ABC_Strong", "ABC_Weak", "ABC_Safe"]


class ABC_Weak(metaclass=ABCMeta_Weak):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance. It follows weak abstraction in the meaning, that
    it is enough to implement the abstarcts in the instance, they impose no
    restriction on the inherited class itself. Therefore, error only occurs at
    runtime.
    
    Notes
    -----
    This class provides nothing over the default behaviour offered by Python's
    standard library, but it is equipped with some useful machinery for
    subclasses that do. All other metaclasses of this module are subclasses of this
    class.
    
    See also
    --------
    :class:`~sigmaepsilon.core.meta.ABCMeta_Weak`

    Examples
    --------
    Create an abstract class with an abstract instance method.

    >>> class ABC_Parent_Weak(ABC_Weak):
    >>>     @abstractmethod
    >>>     def abc_method_parent(self):
    >>>         pass

    Now if we subclass our parent class without implementing `abc_method_parent`,
    nothing happens at runtime (when the class is created).

    >>> class ABC_Child_Weak(ABC_Parent_Weak):
    >>>     @abstractmethod
    >>>     def abc_method_child(self):
    >>>         pass

    If we create an instance with implementing the abstract classes prescribed,
    it works fine.

    >>> class MyClass(ABC_Parent_Weak):
    >>>     def abc_method_child(self):
    >>>         pass
    ...
    >>>     def abc_method_parent(self):
    >>>         pass
    ...
    >>> foo = MyClass()

    If miss any of the required implementations, we can see a `TypeError`:

    >>> class MyClass(ABC_Parent_Weak):
    >>>     def abc_method_child(self):
    >>>         pass
    ...
    >>> foo = MyClass()
    Traceback (most recent call last):
        ...
    TypeError: Can't instantiate abstract class MyClass with abstract methods abc_method_parent
    """
    __slots__ = ()


class ABC_Strong(metaclass=ABCMeta_Strong):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance. It follows strong abstraction in the meaning, that
    an abstract function of any of the base classes must be implemented or
    delayed with the use of another @abstractmethod decorator.
    Contrary to the weak metaclass, error occurs when the class is created.
    
    See also
    --------
    :class:`~sigmaepsilon.core.meta.ABCMeta_Strong`

    Examples
    --------
    Create an abstract class with abstract instance methods.

    >>> class MyParentAbstractClass(ABC_Strong):
    >>>     @abstractmethod
    >>>     def my_abstract_method_A(self):
    >>>         pass
    ...
    >>>     @abstractmethod
    >>>     def my_abstract_method_B(self):
    >>>         pass

    Now if we subclass our parent class, we can either provide an implementation
    of our abstract methods, or delay the implementation by using the `abstractmethod`
    decorator again.

    >>> class MyChildAbstractClass(MyParentAbstractClass):
    >>>     @abstractmethod
    >>>     def my_abstract_method_A(self):
    >>>         # this is delayed
    >>>         pass
    ...
    >>>     def my_abstract_method_B(self):
    >>>         pass

    In every other case, we will se an error when the class is created.

    >>> class MyChildAbstractClass(MyParentAbstractClass):
    >>>     @abstractmethod
    >>>     def my_abstract_method_A(self):
    >>>         pass
    Traceback (most recent call last):
        ...
    TypeError: Can't create abstract class MyChildAbstractClas! MyChildAbstractClas must implement abstract method my_abstract_method_B of class MyParentAbstractClass.
    """
    __slots__ = ()


class ABC_Safe(metaclass=ABCMeta_Safe):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance. Throws a `TypeError`
    if a method tries to shadow a method in any of the base
    classes.
    
    See also
    --------
    :class:`~sigmaepsilon.core.meta.ABCMeta_Safe`

    Examples
    --------
    Create an abstract class with an instance method.

    >>> class ABC_Parent_Safe(ABC_Safe):
    >>>     def funcParentSafe(self):
    >>>         pass

    Now, if we try to overload any implementation in the parent class, we
    got a `TypeError` when the class is created.

    >>> class ABC_Child_Safe(ABC_Parent_Safe):
    >>>     def funcParentSafe(self):
    >>>         pass
    Traceback (most recent call last):
        ...
    TypeError: Can't create abstract class ABC_Child_Safe! Method funcParentSafe is already implemented in class ABC_Parent_Safe.
    """
    __slots__ = ()




# -*- coding: utf-8 -*-
import unittest, os
from os.path import dirname, abspath

import sigmaepsilon.core as sc
from sigmaepsilon.core.config import (
    find_pyproject_toml,
    load_pyproject_config,
    namespace_package_name,
)
from sigmaepsilon.core.testing import SigmaEpsilonTestCase


class TestConfig(SigmaEpsilonTestCase):
    def test_config(self):
        start_dir = os.getcwd()
        pyproject_toml_path = find_pyproject_toml(start_dir, max_depth=0)
        config = load_pyproject_config(filepath=pyproject_toml_path)
        self.assertTrue(isinstance(config, dict))
        config = load_pyproject_config(filepath=pyproject_toml_path, section="project")
        self.assertTrue(isinstance(config, dict))
        config = load_pyproject_config(
            filepath=pyproject_toml_path, section=["project", "build-system"]
        )
        self.assertTrue(isinstance(config, list))
        self.assertTrue(len(config) == 2)
        self.assertTrue(isinstance(config[0], dict))
        self.assertTrue(isinstance(config[1], dict))
        self.assertFailsProperly(
            TypeError, find_pyproject_toml, start_dir, max_depth="0"
        )
        self.assertFailsProperly(
            ValueError, find_pyproject_toml, start_dir, max_depth=-1
        )
        self.assertFailsProperly(TypeError, find_pyproject_toml, 1, max_depth=-1)
        
        config = load_pyproject_config(filepath=pyproject_toml_path, section="project")
        package_name = namespace_package_name(dirname(abspath(sc.__file__)), 10)
        self.assertEqual(package_name, config["name"])
        
        self.assertFailsProperly(
            TypeError, namespace_package_name, start_dir, max_depth="0"
        )
        self.assertFailsProperly(
            ValueError, namespace_package_name, start_dir, max_depth=-1
        )
        self.assertFailsProperly(TypeError, namespace_package_name, 1, max_depth=-1)
        

if __name__ == "__main__":
    unittest.main()



import os
from os.path import dirname, abspath
import appdirs
import warnings
from typing import Optional
from importlib.metadata import metadata

from .wrapping import Wrapper
from .typing import ishashable, issequence
from .cp import classproperty
from .infix import InfixOperator
from .attr import attributor
from .config import namespace_package_name

__pkg_name__ = namespace_package_name(dirname(abspath(__file__)), 10)
__pkg_metadata__ = metadata(__pkg_name__)
__version__ = __pkg_metadata__["version"]
__description__ = __pkg_metadata__["summary"]
del __pkg_metadata__

# catch annoying numpy/vtk future warning:
warnings.simplefilter(action="ignore", category=FutureWarning)

# If available, a local vtk-data instance will be used for examples
SIGMAEPSILON_DATA_PATH: Optional[str] = None
if "SIGMAEPSILON_DATA_PATH" in os.environ:
    SIGMAEPSILON_DATA_PATH = os.environ["SIGMAEPSILON_DATA_PATH"]
    if not os.path.isdir(SIGMAEPSILON_DATA_PATH):
        warnings.warn(
            f"SIGMAEPSILON_DATA_PATH: {SIGMAEPSILON_DATA_PATH} is an invalid path"
        )
    if not os.path.isdir(os.path.join(SIGMAEPSILON_DATA_PATH, "Data")):
        warnings.warn(
            f"SIGMAEPSILON_DATA_PATH: {os.path.join(SIGMAEPSILON_DATA_PATH, 'Data')} does not exist"
        )

# allow user to override the examples path
if "SIGMAEPSILON_USERDATA_PATH" in os.environ:  # pragma: no cover
    USER_DATA_PATH = os.environ["SIGMAEPSILON_USERDATA_PATH"]
    if not os.path.isdir(USER_DATA_PATH):
        raise FileNotFoundError(
            f"Invalid SIGMAEPSILON_USERDATA_PATH at {USER_DATA_PATH}"
        )
else:
    USER_DATA_PATH = appdirs.user_data_dir("SIGMAEPSILON")
    try:
        # Set up data directory
        os.makedirs(USER_DATA_PATH, exist_ok=True)
    except Exception as e:
        warnings.warn(
            f'Unable to create `SIGMAEPSILON_USERDATA_PATH` at "{USER_DATA_PATH}"\n'
            f"Error: {e}\n\n"
            "Override the default path by setting the environmental variable "
            "`SIGMAEPSILON_USERDATA_PATH` to a writable path."
        )
        USER_DATA_PATH = ""

EXAMPLES_PATH = os.path.join(USER_DATA_PATH, "examples")
try:
    os.makedirs(EXAMPLES_PATH, exist_ok=True)
except Exception as e:
    warnings.warn(
        f'Unable to create `EXAMPLES_PATH` at "{EXAMPLES_PATH}"\n'
        f"Error: {e}\n\n"
        "Override the default path by setting the environmental variable "
        "`SIGMAEPSILON_USERDATA_PATH` to a writable path."
    )
    EXAMPLES_PATH = ""


# Set a parameter to control default print format for floats outside
# of the plotter
FLOAT_FORMAT = "{:.3e}"


__all__ = [
    "Wrapper",
    "ishashable",
    "issequence",
    "classproperty",
    "InfixOperator",
    "attributor",
]



# -*- coding: utf-8 -*-
import unittest

from sigmaepsilon.core.misc import get_index_suffix


class TestIndexSuffix(unittest.TestCase):

    def test_1st(self):
        self.assertEqual(get_index_suffix(1), 'st')

    def test_2nd(self):
        self.assertEqual(get_index_suffix(2), 'nd')

    def test_3rd(self):
        self.assertEqual(get_index_suffix(3), 'rd')

    def test_4th(self):
        self.assertEqual(get_index_suffix(4), 'th')

    def test_11th(self):
        self.assertEqual(get_index_suffix(11), 'th')

    def test_12th(self):
        self.assertEqual(get_index_suffix(12), 'th')

    def test_13th(self):
        self.assertEqual(get_index_suffix(13), 'th')

    def test_21st(self):
        self.assertEqual(get_index_suffix(21), 'st')

    def test_22nd(self):
        self.assertEqual(get_index_suffix(22), 'nd')

    def test_23rd(self):
        self.assertEqual(get_index_suffix(23), 'rd')

    def test_24th(self):
        self.assertEqual(get_index_suffix(24), 'th')
        

if __name__ == '__main__':
    unittest.main()


import toml
import os
from typing import Union, Iterable


def find_pyproject_toml(start_dir: str = None, max_depth: int = 10) -> Union[str, None]:
    """
    Returns the path of the pyproject.toml file of the project.

    Parameters
    ----------
    start_dir: str, Optional
        The dictionary to start with. If not provided, the search starts in the current
        working directory. Default is None.
    max_depth: int, Optional
        The maximum folderwise distance from the starting directory. If provided,
        it must be an integer >= 0. Default is 10.
    """
    if start_dir is None:
        start_dir = os.getcwd()

    if not isinstance(start_dir, str):
        raise TypeError("start_dir must be a string")

    current_dir = start_dir
    depth = 0

    if not isinstance(max_depth, int):
        raise TypeError("max_depth must be an integer")

    if not max_depth >= 0:
        raise ValueError("max_depth must be a positive integer")

    while depth <= max_depth:
        pyproject_path = os.path.join(current_dir, "pyproject.toml")

        if os.path.isfile(pyproject_path):
            return pyproject_path

        parent_dir = os.path.dirname(current_dir)

        current_dir = parent_dir
        depth += 1

    return None


def load_pyproject_config(
    filepath: str = None, section: Union[str, Iterable[str]] = None
) -> Union[dict, Iterable[dict]]:
    """
    Returns the contents of the project configuration file (pyproject.toml)
    or sections of it.

    Parameters
    ----------
    filepath: str, Optional
        The path of the config file. If not specified, an attempt is made to locate it.
        Default is None.
    section: Union[str, Iterable[str]], Optional
        One or more section of the config file. If not specified, the entire content is returned.
        Default is None.
    """
    if not filepath:
        filepath = find_pyproject_toml()

    with open(filepath, "r") as f:
        config_toml = toml.load(f)

    if section:
        if isinstance(section, str):
            config = config_toml.get(section, {})
        else:
            config = [config_toml.get(s, {}) for s in section]
    else:
        config = config_toml

    return config


def namespace_package_name(start_dir: str = None, max_depth: int = 10) -> str:
    """
    Returns the name of the SigmaEpsilon namespace package.
    
    Parameters
    ----------
    start_dir: str, Optional
        The dictionary to start with. If not provided, the search starts in the current
        working directory. Default is None.
    max_depth: int, Optional
        The maximum folderwise distance from the starting directory. If provided,
        it must be an integer >= 0. Default is 10.
    """
    if start_dir is None:
        start_dir = os.getcwd()

    if not isinstance(start_dir, str):
        raise TypeError("start_dir must be a string")
    
    current_dir = start_dir
    depth = 0

    if not isinstance(max_depth, int):
        raise TypeError("max_depth must be an integer")

    if not max_depth >= 0:
        raise ValueError("max_depth must be a positive integer")

    while depth <= max_depth:        
        parent_dir_path, current_dir_name = os.path.split(current_dir)
        parent_dir_name = os.path.split(parent_dir_path)[-1]
        
        if os.path.isdir(current_dir) and parent_dir_name=="sigmaepsilon":
            return ".".join([parent_dir_name, current_dir_name])

        parent_dir = os.path.dirname(current_dir)

        current_dir = parent_dir
        depth += 1

    return None



# -*- coding: utf-8 -*-
import unittest

from sigmaepsilon.core.colors import RGB
from sigmaepsilon.core.testing import SigmaEpsilonTestCase


class TestColors(SigmaEpsilonTestCase):
    def test_colors(self):
        RGB(240, 248, 255).hex_format()
       

if __name__ == "__main__":
    unittest.main()


"""
Class property borrowed from the astropy package.
"""
import functools
import threading

_NotFound = object()

__all__ = ["classproperty"]

# TODO: This can still be made to work for setters by implementing an
# accompanying metaclass that supports it; we just don't need that right
# this moment
class classproperty(property):
    """
    Similar to `property`, but allows class-level properties.  That is,
    a property whose getter is like a `classmethod`.
    The wrapped method may explicitly use the `classmethod` decorator (which
    must become before this decorator), or the `classmethod` may be omitted
    (it is implicit through use of this decorator).
    
    .. note::
        classproperty only works for *read-only* properties.  It does not
        currently allow writeable/deletable properties, due to subtleties of how
        Python descriptors work.  In order to implement such properties on a class,
        a metaclass for that class must be implemented.
        
    Parameters
    ----------
    fget: callable
        The function that computes the value of this property (in particular,
        the function when this is used as a decorator) a la `property`.
    doc: str, optional
        The docstring for the property--by default inherited from the getter
        function.
    lazy: bool, optional
        If True, caches the value returned by the first call to the getter
        function, so that it is only called once (used for lazy evaluation
        of an attribute).  This is analogous to `lazyproperty`.  The ``lazy``
        argument can also be used when `classproperty` is used as a decorator
        (see the third example below).  When used in the decorator syntax this
        *must* be passed in as a keyword argument.
        
    Examples
    --------
    ::
        >>> class Foo:
        ...     _bar_internal = 1
        ...     @classproperty
        ...     def bar(cls):
        ...         return cls._bar_internal + 1
        ...
        >>> Foo.bar
        2
        >>> foo_instance = Foo()
        >>> foo_instance.bar
        2
        >>> foo_instance._bar_internal = 2
        >>> foo_instance.bar  # Ignores instance attributes
        2
    As previously noted, a `classproperty` is limited to implementing
    read-only attributes::
        >>> class Foo:
        ...     _bar_internal = 1
        ...     @classproperty
        ...     def bar(cls):
        ...         return cls._bar_internal
        ...     @bar.setter
        ...     def bar(cls, value):
        ...         cls._bar_internal = value
        ...
        Traceback (most recent call last):
        ...
        NotImplementedError: classproperty can only be read-only; use a
        metaclass to implement modifiable class-level properties
    When the ``lazy`` option is used, the getter is only called once::
        >>> class Foo:
        ...     @classproperty(lazy=True)
        ...     def bar(cls):
        ...         print("Performing complicated calculation")
        ...         return 1
        ...
        >>> Foo.bar
        Performing complicated calculation
        1
        >>> Foo.bar
        1
    If a subclass inherits a lazy `classproperty` the property is still
    re-evaluated for the subclass::
        >>> class FooSub(Foo):
        ...     pass
        ...
        >>> FooSub.bar
        Performing complicated calculation
        1
        >>> FooSub.bar
        1
    """

    def __new__(cls, fget=None, doc=None, lazy=False):
        if fget is None:
            # Being used as a decorator--return a wrapper that implements
            # decorator syntax
            def wrapper(func):
                return cls(func, lazy=lazy)

            return wrapper

        return super().__new__(cls)

    def __init__(self, fget, doc=None, lazy=False):
        self._lazy = lazy
        if lazy:
            self._lock = threading.RLock()  # Protects _cache
            self._cache = {}
        fget = self._wrap_fget(fget)

        super().__init__(fget=fget, doc=doc)

        # There is a buglet in Python where self.__doc__ doesn't
        # get set properly on instances of property subclasses if
        # the doc argument was used rather than taking the docstring
        # from fget
        # Related Python issue: https://bugs.python.org/issue24766
        if doc is not None:
            self.__doc__ = doc

    def __get__(self, obj, objtype):
        if self._lazy:
            val = self._cache.get(objtype, _NotFound)
            if val is _NotFound:
                with self._lock:
                    # Check if another thread initialised before we locked.
                    val = self._cache.get(objtype, _NotFound)
                    if val is _NotFound:
                        val = self.fget.__wrapped__(objtype)
                        self._cache[objtype] = val
        else:
            # The base property.__get__ will just return self here;
            # instead we pass objtype through to the original wrapped
            # function (which takes the class as its sole argument)
            val = self.fget.__wrapped__(objtype)
        return val

    def getter(self, fget):
        return super().getter(self._wrap_fget(fget))

    def setter(self, fset):
        raise NotImplementedError(
            "classproperty can only be read-only; use a metaclass to "
            "implement modifiable class-level properties"
        )

    def deleter(self, fdel):
        raise NotImplementedError(
            "classproperty can only be read-only; use a metaclass to "
            "implement modifiable class-level properties"
        )

    @staticmethod
    def _wrap_fget(orig_fget):
        if isinstance(orig_fget, classmethod):
            orig_fget = orig_fget.__func__

        # Using stock functools.wraps instead of the fancier version
        # found later in this module, which is overkill for this purpose

        @functools.wraps(orig_fget)
        def fget(obj):
            return orig_fget(obj.__class__)

        return fget



# -*- coding: utf-8 -*-
from typing import Union

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

__all__ = ["alphapet", "ordrange", "latinrange", "greekrange", "urange", "arabicrange"]


def alphabet(abctype: str = "latin", start: Union[str, int] = None) -> Iterable:
    """
    Generator function that yields characters from different alphabets based on
    the specified `abctype` and `start` parameters.

    Parameters
    ----------
    abctype: str, Optional
        The tipe of the alphabet. Accepted values are 'ord' or 'o' for ASCII ordinal values,
        'latin' or 'l' for Latin, 'u' for Unicode and 'greek' or 'g' for Greek letters.
        Default is 'latin'.
    start: Union[str, int], Optional
        The code of the character to start with. Default is `None`, in which case
        a character is automatically selected as the first letter of the alphabet.

    Yields
    ------
    string
        A character from the specified alphabet.

    Examples
    --------
    To generate 5 greek characters:
    >>> from sigmaepsilon.core.alphabet import alphabet
    >>> f = alphabet('greek')
    >>> characters = list(map(alphabet('greek'), range(5)))
    """
    if abctype in ("ord", "o"):
        start = 0 if start is None else start
    elif abctype in ("latin", "l"):
        start = ord("a") if start is None else start
    elif abctype == "u":
        start = ord("\u0000") if start is None else start
    elif abctype in ("greek", "g"):
        start = ord("\u03b1") if start is None else start
    else:
        raise ValueError("Invalid alphabet.")
    start = ord(start) if isinstance(start, str) else start
    while True:
        yield chr(start)
        start += 1


def ordrange(N: int = 1, **kwargs) -> Iterable:
    start = kwargs.pop("start", 0)
    if isinstance(start, str):
        start = ord(start)
    stop = kwargs.pop("stop", None)
    if stop is None or stop == start:
        stop = start + N
    return [chr(c) for c in range(start, stop)]


def latinrange(N: int = 1, **kwargs) -> Iterable:
    start = kwargs.pop("start", 97)
    stop = kwargs.pop("stop", None)
    return ordrange(N, start=start, stop=stop)


def urange(N: int = 1, **kwargs) -> Iterable:
    start = kwargs.pop("start", "\u0000")
    stop = kwargs.pop("stop", None)
    if stop is None:
        stop = start
    return ordrange(N, start=ord(start), stop=ord(stop))


def greekrange(N: int = 1) -> Iterable:
    return urange(N, start="\u03b1")


def arabicrange(N: int = 1, **kwargs) -> Iterable:
    start = kwargs.pop("start", 0)
    stop = kwargs.pop("stop", None)
    if stop is None or stop == start:
        stop = start + N
    return [str(c) for c in range(start, stop)]



"""
Warnings used in SigmaEpsilon projects.
"""

__all__ = ["SigmaEpsilonWarning", "SigmaEpsilonPerformanceWarning"]


class SigmaEpsilonWarning(Warning):
    """
    Base class for warnings in SigmaEpsilon projects.
    """
    ...


class SigmaEpsilonPerformanceWarning(SigmaEpsilonWarning):
    """
    Performance warning in SigmaEpsilon projects.
    """
    def __init__(self, message: str):
        pre = "SigmaEpsilon Performance Warning: "
        self.message = pre + message


# -*- coding: utf-8 -*-
from typing import Any

__all__ = ["Wrapper", "wrapper", "customwrapper", "wrap"]

NoneType = type(None)


class Wrapper:
    """
    Wrapper base class that makes it easy (and safe) to extend other objects.
    Based on the provided arguments at initialization, the wrapper either
    
    (a) wraps an existing object at object creation provided as a keyword
        argument with `Wrapper.wrapkey`
    (b) wraps an existing object at object creation if it is a positional
        argument and an instance of `Wrapper.wraptype`
    (b) wraps the object Wrapper.wraptype(*args, **kwargs) if
        `Wrapper.wraptype` is not `None`
        
    The attributes and methods of the wrapped instance are all accessible
    through the wrapper.
    
    Using a wrapper is a good idea if you want to easily extend the functionality
    provided by a class of an external library without having to worry about shadowing
    an important method and thus risking to break the behaviour of the wrapped object.
        
    Examples
    --------
    >>> import numpy as np
    >>>
    >>> arr = np.eye(3)
    >>> wrapper = Wrapper(wrap=arr)
    
    Now the wrapped NumPy array is accessible as `wrapper.wrapped`.
    
    A wrapper class can be used to extend the behaviour:
    
    >>> MyWrapper(Wrapper):
    >>>     wraptype = np.ndarray
    >>>     
    >>>     def invert() -> None:
    >>>         self.wrapped = np.linalg.inv(self.wrapped)
    
    With this solution, you don't have to worry about shadowing an existing
    implementation of NumPy arrays. Not that NumPy arrays might not have a method
    called `invert` at the time of implementing the class `MyWrapper`, but it might
    change in the future. You can even check that internally and get notified: 
    
    >>> import warnings
    >>>
    >>> MyWrapper(Wrapper):
    >>>     wraptype = np.ndarray
    >>>     
    >>>     def invert() -> None:
    >>>         if hasattr(self.wrapped, "invert"):
    >>>             warnings.warn("'invert' already exists in the object")
    >>>         self.wrapped = np.linalg.inv(self.wrapped)
    
    Then, to wrap a NumPy array, you can do this (since the `wraptype` attribute is set,
    the MyWrapper class is going to catch the object to wrap as a positional argument):
    
    >>> wrapper = MyWrapper(arr)
    """
    wrapkey: str = "wrap"
    wraptype: Any = NoneType

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._wrapped = None
        self.wrap(kwargs.get(self.wrapkey, None))

        if self._wrapped is None and self.wraptype is not NoneType:
            for arg in args:
                if isinstance(arg, self.wraptype):
                    self._wrapped = arg
                    break

            if self._wrapped is None:
                try:
                    if self.wraptype is not NoneType:
                        self._wrapped = self.wraptype(*args, **kwargs)
                except Exception:
                    raise ValueError(
                        "Wrapped class '{}' cannot be "
                        + "initiated with these "
                        + "arguments".format(self.wraptype.__name__)
                    )
        else:
            if self.wraptype is not NoneType:
                assert isinstance(
                    self._wrapped, self.wraptype
                ), "Wrong type, unable to wrap object : {}".format(self._wrapped)

    @property
    def wrapped(self):
        """Retruns the wrapped object."""
        return self._wrapped

    def wrap(self, obj: Any=None) -> Any:
        """Wraps the provided object and returns the wrapper instance."""
        if self.wraptype is not NoneType:
            if isinstance(obj, self.wraptype):
                self._wrapped = obj
        else:
            self._wrapped = obj
        return self

    def wraps(self):
        """Returns `True` if the instance wraps something or `False` if it doesn't."""
        return self._wrapped is not None

    def wrapped_obj(self):
        return self._wrapped

    def __hasattr__(self, attr):
        return any([attr in self.__dict__, attr in self._wrapped.__dict__])

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)
        try:
            return getattr(self._wrapped, attr)
        except Exception:
            raise AttributeError(
                "'{}' object has no attribute \
                called {}".format(
                    self.__class__.__name__, attr
                )
            )

    def __getitem__(self, index):
        try:
            return super().__getitem__(index)
        except Exception:
            try:
                return self._wrapped.__getitem__(index)
            except Exception:
                raise TypeError(
                    "'{}' object is not "
                    "subscriptable".format(self.__class__.__name__)
                )

    def __setitem__(self, index, value):
        try:
            return super().__setitem__(index, value)
        except Exception:
            try:
                return self._wrapped.__setitem__(index, value)
            except Exception:
                raise TypeError(
                    "'{}' object does not support "
                    "item assignment".format(self.__class__.__name__)
                )


def customwrapper(*, wrapkey:str="wrap", wraptype: Any=NoneType) -> Wrapper:
    """
    A factory function that returns a class decorator turning a class type 
    into a wrapper type, that either
    
    (a) wraps an existing object at object creation provided as a keyword
        argument with wrapkey
    (b) wraps an existing object at object creation if it is a positional
        argument and an instance of wraptype
    (b) wraps the object wraptype(*args, **kwargs)
    
    See also
    --------
    :class:`~sigmaepsilon.core.wrapping.Wrapper`
    
    Examples
    --------
    Take this example from the :class:`~sigmaepsilon.core.wrapping.Wrapper` class:
    
    >>> MyWrapper(Wrapper):
    >>>     wraptype = np.ndarray
    >>>
    >>>     def invert() -> None:
    >>>         self.wrapped = np.linalg.inv(self.wrapped)
    
    An equivalent implementation of this is the following:
    
    >>> @customwrapper(wraptype=np.ndarray)
    >>> MyWrapper:
    >>>     def invert() -> None:
    >>>         self.wrapped = np.linalg.inv(self.wrapped)
    
    Notice how the class `MyWrapper` is not inherited from the `Wrapper` class.
    """

    class BaseWrapperType(Wrapper):
        ...

    BaseWrapperType.wrapkey = wrapkey
    BaseWrapperType.wraptype = wraptype

    def wrapper(BaseType):
        class WrapperType(BaseWrapperType, BaseType):
            basetype = BaseType

        return WrapperType

    return wrapper


def wrapper(BaseType: Any) -> Wrapper:
    """
    Simple class decorator that turns a type into a wrapper with default
    behaviour.
    
    Notes
    -----
    This is the same as using the :func:`~sigmaepsilon.core.wrapping.customwrapper`
    with the default values.
    
    See also
    --------
    :class:`~sigmaepsilon.core.wrapping.Wrapper`
    :func:`~sigmaepsilon.core.wrapping.customwrapper`
    
    Example
    -------
    >>> @wrapper
    >>> MyWrapper:
    >>>     def invert() -> None:
    >>>         self.wrapped = np.linalg.inv(self.wrapped)
    """
    class WrapperType(Wrapper, BaseType):
        basetype = BaseType

    return WrapperType


def wrap(obj: object) -> Wrapper:
    """
    Wraps an object and returns the wrapper.
    
    See also
    --------
    :class:`~sigmaepsilon.core.wrapping.Wrapper`
    
    Example
    -------
    >>> import numpy as np
    >>> wrapper = wrap(np.eye(3))
    """
    return Wrapper(wrap=obj)




def get_index_suffix(index: int) -> str:
    """
    Returns the proper suffix for an integer.
    """
    if index % 100 in {11, 12, 13}:  # Special case for 11th, 12th, and 13th
        suffix = "th"
    else:
        last_digit = index % 10
        if last_digit == 1:
            suffix = "st"
        elif last_digit == 2:
            suffix = "nd"
        elif last_digit == 3:
            suffix = "rd"
        else:
            suffix = "th"
    return suffix


# -*- coding: utf-8 -*-
import unittest

from sigmaepsilon.core.cp import classproperty


class TestProperty(unittest.TestCase):

    def test_class_property(self):

        class TestClasss:

            @classproperty
            def prop(cls):
                return 1

        self.assertEqual(TestClasss.prop, 1)


if __name__ == "__main__":
    unittest.main()
